/**
 *
 * $Id: DragOverS.c,v 1.11 1997/02/19 02:50:21 miers Exp $
 *
 * derived from Xt Vendor class.c 
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 * Copyright 1989 Massachusetts Institute of Technology
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

static char rcsid[] = "$Id: DragOverS.c,v 1.11 1997/02/19 02:50:21 miers Exp $";

#include <LTconfig.h>
#include <XmI/XmI.h>

#include <Xm/XmP.h>
#include <Xm/VendorSP.h>
#include <Xm/VendorSEP.h>
#include <Xm/DragOverSP.h>
#include <Xm/DropSMgrP.h>
#include <Xm/DragCP.h>
#include <Xm/ScreenP.h>

#include <XmI/DebugUtil.h>

static void class_initialize();
static void class_part_initialize(WidgetClass wclass);
static void initialize(Widget request, Widget new_w, ArgList args, Cardinal *num_args);
static Boolean set_values(Widget old, Widget request, Widget new_w, ArgList args, Cardinal *num_args);
static void expose(Widget w, XEvent *event, Region region);
static void destroy(Widget w);
static void change_managed(Widget w);

static XtResource resources[] = {
    {
	XmNoverrideRedirect, XmCOverrideRedirect, XmRBoolean,
	sizeof(Boolean), XtOffsetOf(XmDragOverShellRec, shell.override_redirect),
	XmRImmediate, (XtPointer)True
    },
    {
	XmNhotX, XmCHot, XmRPosition,
	sizeof(Position), XtOffsetOf(XmDragOverShellRec, drag.hotX),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNhotY, XmCHot, XmRPosition,
	sizeof(Position), XtOffsetOf(XmDragOverShellRec, drag.hotY),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNdragOverMode, XmCDragOverMode, XmRUnsignedChar,
	sizeof(unsigned char), XtOffsetOf(XmDragOverShellRec, drag.mode),
	XmRImmediate, (XtPointer)XmCURSOR
    }
};

static XmBaseClassExtRec _XmDragOverSCoreClassExtRec = {
    /* next_extension            */ NULL,
    /* record_type               */ NULLQUARK,                             
    /* version                   */ XmBaseClassExtVersion,
    /* size                      */ sizeof(XmBaseClassExtRec),
    /* initialize_prehook        */ XmInheritInitializePrehook,
    /* set_values_prehook        */ XmInheritSetValuesPrehook,
    /* initialize_posthook       */ XmInheritInitializePosthook,
    /* set_values_posthook       */ XmInheritSetValuesPosthook,
    /* secondary_object_class    */ XmInheritClass,
    /* secondary_object_create   */ XmInheritSecObjectCreate,
    /* get_secondary_resources   */ XmInheritGetSecResData,
    /* fast_subclass             */ { 0 },
    /* get_values_prehook        */ XmInheritGetValuesPrehook,
    /* get_values_posthook       */ XmInheritGetValuesPosthook,
    /* class_part_init_prehook   */ XmInheritClassPartInitPrehook,
    /* class_part_init_posthook  */ XmInheritClassPartInitPosthook,
    /* ext_resources             */ NULL,
    /* compiled_ext_resources    */ NULL,
    /* num_ext_resources         */ 0,
    /* use_sub_resources         */ False,
    /* widget_navigable          */ XmInheritWidgetNavigable,
    /* focus_change              */ XmInheritFocusChange,
    /* wrapper_data              */ NULL
};

XmDragOverShellClassRec xmDragOverShellClassRec = {
    /* Core class part */
    {
	/* superclass            */ (WidgetClass) &vendorShellClassRec,
        /* class_name            */ "XmDragOverShell",
        /* size                  */ sizeof(XmDragOverShellRec),
        /* class_initialize      */ class_initialize,
        /* class_part_initialize */ class_part_initialize,
        /* class_inited          */ False,
        /* initialize            */ initialize,
      	/* initialize_hook       */ NULL,		
      	/* realize               */ XtInheritRealize,
      	/* actions               */ NULL,
      	/* num_actions           */ 0,
      	/* resources             */ resources,
      	/* resource_count        */ XtNumber(resources),
      	/* xrm_class             */ NULLQUARK,
      	/* compress_motion       */ False,
      	/* compress_exposure     */ True,
      	/* compress_enterleave   */ False,
      	/* visible_interest      */ False,
      	/* destroy               */ destroy,
      	/* resize                */ XtInheritResize,
      	/* expose                */ expose,
      	/* set_values            */ set_values,
      	/* set_values_hook       */ NULL,
      	/* set_values_almost     */ XtInheritSetValuesAlmost,  
      	/* get_values_hook       */ NULL,
      	/* accept_focus          */ NULL,
      	/* intrinsics version    */ XtVersion,
      	/* callback offsets      */ NULL,
      	/* tm_table              */ NULL,
      	/* query_geometry        */ XtInheritQueryGeometry,
      	/* display_accelerator   */ NULL,
      	/* extension             */ NULL /*(XtPointer)&_XmDragOverSCoreClassExtRec */
    },
    /* Composite Class Part */
    {
	/* geometry_manager */ XtInheritGeometryManager,
        /* change_managed   */ change_managed,
        /* insert_child     */ XtInheritInsertChild,
        /* delete_child     */ XtInheritDeleteChild,
        /* extension        */ NULL
    },
    /* Shell Class Part */
    {
	/* extension */ NULL
    },
    /* WMShell Class Part*/
    {
	/* extension */ NULL
    },
    /* Vendor Shell Class */
    {
	/* extension */	NULL
    },
    /* XmDragOverShell Class Part */
    {
	/* extension */ NULL
    }
};

WidgetClass xmDragOverShellWidgetClass = (WidgetClass)(&xmDragOverShellClassRec);

/*
 * Things to remember:
 *  o We're a popup shell.
 *  o Our parent is a DragContext.
 *  o blendModel comes from DragContext, and is one of:
 *      XmBLEND_ALL, XmBLEND_STATE_SOURCE, XmBLEND_JUST_SOURCE, XmBLEND_NONE
 *  o mode and activeMode is ours, and is (I think) one of:
 *	XmWINDOW, XmPIXMAP, XmCURSOR
 *  o provides the canvas for animation effects
 */

static void class_initialize()
{
    _XmDragOverSCoreClassExtRec.record_type = XmQmotif;
}

static void class_part_initialize(WidgetClass widget_class)
{
    _XmFastSubclassInit(widget_class, XmDRAG_OVER_SHELL_BIT);
}


static void 
initialize(Widget request, 
	   Widget new_w, 
	   ArgList args, 
	   Cardinal *num_args)
{
    XtWidth(new_w) = 0;
    XtHeight(new_w) = 0;

    DO_ActiveMode(new_w) = XmCURSOR;

    DO_InitialX(new_w) = DO_HotX(new_w);
    DO_InitialY(new_w) = DO_HotY(new_w);

    DO_CursorBlend(new_w).sourceIcon = NULL;
    DO_CursorBlend(new_w).sourceX = 0;
    DO_CursorBlend(new_w).sourceY = 0;
    DO_CursorBlend(new_w).mixedIcon = NULL;
    DO_CursorBlend(new_w).gc = NULL;

    DO_RootBlend(new_w).sourceIcon = NULL;
    DO_RootBlend(new_w).sourceX = 0;
    DO_RootBlend(new_w).sourceY = 0;
    DO_RootBlend(new_w).mixedIcon = NULL;
    DO_RootBlend(new_w).gc = NULL;

    DO_NCCursor(new_w) = None;
    DO_ActiveCursor(new_w) = None;

    DO_Backing(new_w).pixmap = XmUNSPECIFIED_PIXMAP;
    DO_Backing(new_w).x = 0;
    DO_Backing(new_w).y = 0;

    DO_TmpPix(new_w) = XmUNSPECIFIED_PIXMAP;
    DO_TmpBit(new_w) = XmUNSPECIFIED_PIXMAP;

    DO_IsVisible(new_w) = False;

    /* grabs on this here would be a really bad idea */
    XtRemoveAllCallbacks(new_w, XmNpopupCallback);
    XtRemoveAllCallbacks(new_w, XmNpopdownCallback);

    /* gotta build the mixed cursor */
    _XmDragOverChange(new_w, XmINVALID_DROP_SITE);
}

static void 
destroy(Widget w)
{
}

static Boolean 
set_values(Widget old, 
	   Widget request, 
	   Widget new_w, 
	   ArgList args, 
	   Cardinal *num_args)
{
    return False;
}

static void
expose(Widget w, XEvent *event, Region region)
{
}

static void 
change_managed(Widget w)
{
}

/*
 * compute the position of the op and state cursors (obj1) relative to
 * the source cursor (obj2)
 */
static void
compute_pos(Widget w, XmDragIconObject obj1, XmDragIconObject obj2,
	    Position *x, Position *y)
{
    Widget par;	/* This had BETTER be a drag context */
    Widget dos;

    switch (DI_Attachment(obj1)) {
    default:
	_XmWarning(w, "Invalid attachment for DragIcon.");
	/* FALL THROUGH */

    case XmATTACH_NORTH_WEST:
	*x = DI_OffsetX(obj1);
	*y = DI_OffsetY(obj1);
	return;

    case XmATTACH_NORTH:
	*x = (DI_Width(obj2) >> 1) + DI_OffsetX(obj1);
	*y = DI_OffsetY(obj1);
	return;

    case XmATTACH_NORTH_EAST:
	*x = DI_Width(obj2) + DI_OffsetX(obj1);
	*y = DI_OffsetY(obj1);
	return;

    case XmATTACH_EAST:
	*x = DI_Width(obj2) + DI_OffsetX(obj1);
	*y = (DI_Height(obj2) >> 1) + DI_OffsetY(obj1);
	return;

    case XmATTACH_SOUTH_EAST:
	*x = DI_Width(obj2) + DI_OffsetX(obj1);
	*y = DI_Height(obj2) + DI_OffsetY(obj1);
	return;

    case XmATTACH_SOUTH:
	*x = (DI_Width(obj2) >> 1) + DI_OffsetX(obj1);
	*y = DI_Height(obj2) + DI_OffsetY(obj1);
	return;

    case XmATTACH_SOUTH_WEST:
	*x = DI_OffsetX(obj1);
	*y = DI_Height(obj2) + DI_OffsetY(obj1);
	return;

    case XmATTACH_WEST:
	*x = DI_OffsetX(obj1);
	*y = (DI_Height(obj2) >> 1) + DI_OffsetY(obj1);
	return;

    case XmATTACH_CENTER:
	*x = (DI_Width(obj2) >> 1) + DI_OffsetX(obj1);
	*y = (DI_Height(obj2) >> 1) + DI_OffsetY(obj1);
	return;

    case XmATTACH_HOT:
	par = XtParent(w);
	if (DC_OrigDragOver(par) != NULL)
	    dos = (Widget)DC_OrigDragOver(par);
	else
	    dos = w;

	if (DO_RootBlend(dos).mixedIcon != NULL) {
	    *x = DI_HotX(DO_RootBlend(dos).mixedIcon) -
			DO_RootBlend(dos).sourceX -
			DI_HotX(obj1);
	    *y = DI_HotY(DO_RootBlend(dos).mixedIcon) -
			DO_RootBlend(dos).sourceY -
			DI_HotY(obj1);
	}
	else if (DO_CursorBlend(dos).mixedIcon != NULL) {
	    *x = DI_HotX(DO_CursorBlend(dos).mixedIcon) -
			DO_CursorBlend(dos).sourceX -
			DI_HotX(obj1);
	    *y = DI_HotY(DO_CursorBlend(dos).mixedIcon) -
			DO_CursorBlend(dos).sourceY -
			DI_HotY(obj1);
	}
	else {
	    /* FIXME */
	}
    }
}

/*
 * compute the size of the combined cursor
 */
static void
compute_size(Widget w,
	     XmDragIconObject src, XmDragIconObject op, XmDragIconObject state,
	     Dimension *wd, Dimension *ht)
{
    Dimension cw = 0, ch = 0;
    Position x = 0, y = 0, sox, soy, stx, sty;

    if (op != NULL) {
	compute_pos(w, op, src, &sox, &soy);
	if (sox < x)
	    x = sox;
	if (soy < y)
	    y = soy;
    }

    if (state != NULL) {
	if (DI_Attachment(state) == XmATTACH_HOT && op != NULL) {
	    stx = sox + DI_HotX(op) - DI_HotX(state);
	    sty = soy + DI_HotY(op) - DI_HotY(state);
	}
	else
	    compute_pos(w, state, src, &stx, &sty);

	if (stx < x)
	    x = stx;
	if (sty < y)
	    y = sty;
    }

    cw = -x;
    ch = -y;
    cw += DI_Width(src);
    ch += DI_Height(src);

    if (op != NULL) {
	sox -= x;
	soy -= y;
	if (sox + DI_Width(op) > cw)
	    cw = sox + DI_Width(op);
	if (soy + DI_Height(op) > ch)
	    ch = soy + DI_Height(op);
    }

    if (state != NULL) {
	stx -= x;
	sty -= y;
	if (stx + DI_Width(state) > cw)
	    cw = stx + DI_Width(state);
	if (sty + DI_Height(state) > ch)
	    ch = sty + DI_Height(state);
    }

    *wd = cw;
    *ht = ch;
}

static Boolean
cursor_fits(Widget w,
	    XmDragIconObject src, XmDragIconObject op, XmDragIconObject state)
{
    Dimension cw, ch, sw, sh;

    if (DI_Depth(src) != 1 || DI_Pixmap(src) == XmUNSPECIFIED_PIXMAP)
	return False;

    compute_size(w, src, op, state, &cw, &ch);

    _XmGetMaxCursorSize(w, &sw, &sh);

    if (cw > sw || ch > sh)
	return False;

    return True;
}

/*
 * blend the icons into a single image.
 */
static void
blend_icons(Widget w, XmDragOverBlend blend,
	    XmDragIconObject src, XmDragIconObject op, XmDragIconObject state)
{
    XGCValues values;
    Display *disp = XtDisplay(w);
    Screen *scr = XtScreen(w);
    XRectangle rect;
    unsigned char bm = DC_BlendModel(XtParent(w));

    if (blend->gc == NULL) {
	values.foreground = 1;
	values.background = 0;
	values.graphics_exposures = False;
	values.fill_style = FillSolid;
	values.function = GXclear;
	blend->gc = XCreateGC(disp, DI_Pixmap(blend->mixedIcon),
			      GCForeground|GCBackground|GCGraphicsExposures|
				GCFillStyle|GCFunction,
			      &values);
    }
    else {
	values.foreground = BlackPixelOfScreen(scr);
	values.background = WhitePixelOfScreen(scr);
	values.graphics_exposures = False;
	values.fill_style = FillSolid;
	values.function = GXclear;
	XChangeGC(disp, blend->gc,
		  GCForeground|GCBackground|GCGraphicsExposures|
		   GCFillStyle|GCFunction,
		  &values);
    }

    rect.x = rect.y = 0;
    rect.width = DI_Width(blend->mixedIcon);
    rect.height = DI_Height(blend->mixedIcon);
    XSetClipRectangles(disp, blend->gc, 0, 0, &rect, 1, Unsorted);

    /* clear it out */
    if (DI_Mask(blend->mixedIcon) != XmUNSPECIFIED_PIXMAP)
        XCopyPlane(disp,
		   DI_Mask(blend->mixedIcon), DI_Mask(blend->mixedIcon),
		   blend->gc,
		   0, 0,
		   DI_Width(blend->mixedIcon), DI_Height(blend->mixedIcon),
		   0, 0, 1);

    XSetFunction(disp, blend->gc, GXor);

    if (DI_Mask(blend->mixedIcon) != XmUNSPECIFIED_PIXMAP &&
	bm != XmBLEND_NONE) {

	/* copy in the masks */
	rect.x = DI_XOffset(src);
	rect.y = DI_YOffset(src);
	rect.width = DI_Width(src);
	rect.height = DI_Height(src);
	XSetClipRectangles(disp, blend->gc, 0, 0, &rect, 1, Unsorted);

	if (DI_Mask(src) != XmUNSPECIFIED_PIXMAP) {
	    XCopyPlane(disp, DI_Mask(src), DI_Mask(blend->mixedIcon),
		       blend->gc, 0, 0, DI_Width(src), DI_Height(src),
		       DI_XOffset(src), DI_YOffset(src), 1);
	}
	else {
	    XSetFunction(disp, blend->gc, GXset);
	    XFillRectangles(disp, DI_Mask(blend->mixedIcon),
			    blend->gc, &rect, 1);
	    XSetFunction(disp, blend->gc, GXor);
	}

	rect.x = DI_XOffset(state);
	rect.y = DI_YOffset(state);
	rect.width = DI_Width(state);
	rect.height = DI_Height(state);
	XSetClipRectangles(disp, blend->gc, 0, 0, &rect, 1, Unsorted);
	if (DI_Mask(state) != XmUNSPECIFIED_PIXMAP &&
	    bm != XmBLEND_JUST_SOURCE) {
	    XCopyPlane(disp, DI_Mask(state), DI_Mask(blend->mixedIcon),
		       blend->gc, 0, 0, DI_Width(state), DI_Height(state),
		       DI_XOffset(state),
		       DI_YOffset(state), 1);
	}
	else if (bm != XmBLEND_JUST_SOURCE) {
	    XSetFunction(disp, blend->gc, GXset);
	    XFillRectangles(disp, DI_Mask(blend->mixedIcon),
			    blend->gc, &rect, 1);
	    XSetFunction(disp, blend->gc, GXor);
	}

	rect.x = DI_XOffset(op);
	rect.y = DI_YOffset(op);
	rect.width = DI_Width(op);
	rect.height = DI_Height(op);
	XSetClipRectangles(disp, blend->gc, 0, 0, &rect, 1, Unsorted);
	if (DI_Mask(op) != XmUNSPECIFIED_PIXMAP &&
	    bm != XmBLEND_JUST_SOURCE && bm != XmBLEND_STATE_SOURCE) {
	    XCopyPlane(disp, DI_Mask(op), DI_Mask(blend->mixedIcon),
		       blend->gc, 0, 0, DI_Width(op), DI_Height(op),
		       DI_XOffset(op),
		       DI_YOffset(op), 1);
	}
	else if (bm != XmBLEND_JUST_SOURCE && bm != XmBLEND_STATE_SOURCE) {
	    XSetFunction(disp, blend->gc, GXset);
	    XFillRectangles(disp, DI_Mask(blend->mixedIcon),
			    blend->gc, &rect, 1);
	    XSetFunction(disp, blend->gc, GXor);
	}
    }

    values.function = GXclear;
    XChangeGC(disp, blend->gc,
	      GCFunction,
	      &values);

    if (DI_Pixmap(blend->mixedIcon) != XmUNSPECIFIED_PIXMAP) {
	XCopyArea(disp,
		  DI_Pixmap(blend->mixedIcon), DI_Pixmap(blend->mixedIcon),
		  blend->gc,
		  0, 0,
		  DI_Width(blend->mixedIcon), DI_Height(blend->mixedIcon),
		  0, 0);
    }

    values.function = GXcopy;
    XChangeGC(disp, blend->gc,
	      GCFunction,
	      &values);

    rect.x = rect.y = 0;
    rect.width = DI_Width(blend->mixedIcon);
    rect.height = DI_Height(blend->mixedIcon);
    XSetClipRectangles(disp, blend->gc, 0, 0, &rect, 1, Unsorted);

    if (bm == XmBLEND_NONE)
	return;

    if (src && DI_Pixmap(src) != XmUNSPECIFIED_PIXMAP) {

	if (DI_Mask(src) != XmUNSPECIFIED_PIXMAP) {
	    XSetClipOrigin(disp, blend->gc,
			   DI_XOffset(src), DI_YOffset(src));
	    XSetClipMask(disp, blend->gc, DI_Mask(src));
	}
	else {
	    rect.x = DI_XOffset(src);
	    rect.y = DI_YOffset(src);
	    rect.width = DI_Width(src);
	    rect.height = DI_Height(src);
	    XSetClipRectangles(disp, blend->gc, 0, 0, &rect, 1, Unsorted);
	}
	if (DI_Depth(src) == DI_Depth(blend->mixedIcon))
	    XCopyArea(disp, DI_Pixmap(src), DI_Pixmap(blend->mixedIcon),
		      blend->gc, 0, 0, DI_Width(src), DI_Height(src),
		      DI_XOffset(src), DI_YOffset(src));
	else
	    XCopyPlane(disp, DI_Pixmap(src), DI_Pixmap(blend->mixedIcon),
		       blend->gc, 0, 0, DI_Width(src), DI_Height(src),
		       DI_XOffset(src), DI_YOffset(src), 1);

    }
    if (state && DI_Pixmap(state) != XmUNSPECIFIED_PIXMAP &&
	bm != XmBLEND_JUST_SOURCE) {

	if (DI_Mask(state) != XmUNSPECIFIED_PIXMAP) {
	    XSetClipOrigin(disp, blend->gc,
			   DI_XOffset(state), DI_YOffset(state));
	    XSetClipMask(disp, blend->gc, DI_Mask(state));
	}
	else
	{
	    rect.x = DI_XOffset(state);
	    rect.y = DI_YOffset(state);
	    rect.width = DI_Width(state);
	    rect.height = DI_Height(state);
	    XSetClipRectangles(disp, blend->gc, 0, 0, &rect, 1, Unsorted);
	}
	if (DI_Depth(state) == DI_Depth(blend->mixedIcon))
	    XCopyArea(disp, DI_Pixmap(state), DI_Pixmap(blend->mixedIcon),
		      blend->gc, 0, 0, DI_Width(state), DI_Height(state),
		      DI_XOffset(state),
		      DI_YOffset(state));
	else
	    XCopyPlane(disp, DI_Pixmap(state), DI_Pixmap(blend->mixedIcon),
		       blend->gc, 0, 0, DI_Width(state), DI_Height(state),
		       DI_XOffset(state),
		       DI_YOffset(state), 1);

    }
    if (op && DI_Pixmap(op) != XmUNSPECIFIED_PIXMAP &&
	bm != XmBLEND_JUST_SOURCE && bm != XmBLEND_STATE_SOURCE) {

	if (DI_Mask(op) != XmUNSPECIFIED_PIXMAP) {
	    XSetClipOrigin(disp, blend->gc,
			   DI_XOffset(op), DI_YOffset(op));
	    XSetClipMask(disp, blend->gc, DI_Mask(op));
	}
	else {
	    rect.x = DI_XOffset(op);
	    rect.y = DI_YOffset(op);
	    rect.width = DI_Width(op);
	    rect.height = DI_Height(op);
	    XSetClipRectangles(disp, blend->gc, 0, 0, &rect, 1, Unsorted);
	}
	if (DI_Depth(op) == DI_Depth(blend->mixedIcon))
	    XCopyArea(disp, DI_Pixmap(op), DI_Pixmap(blend->mixedIcon),
		      blend->gc, 0, 0, DI_Width(op), DI_Height(op),
		      DI_XOffset(op),
		      DI_YOffset(op));
	else
	    XCopyPlane(disp, DI_Pixmap(op), DI_Pixmap(blend->mixedIcon),
		       blend->gc, 0, 0, DI_Width(op), DI_Height(op),
		       DI_XOffset(op),
		       DI_YOffset(op), 1);
    }
}

/*
 * combine the cursors
 */
static void
make_soup(Widget w,
	  XmDragIconObject src, XmDragIconObject op, XmDragIconObject state,
	  XmDragOverBlend blend, Boolean check_fit)
{
    XmScreen s = (XmScreen)XmGetXmScreen(XtScreen(w));
    Position sox = 0, soy = 0;
    Position stx = 0, sty = 0;
    Dimension cw, ch;
    Dimension lw, lh;
    Dimension sw, sh;
    Position x = 0, y = 0;
    Pixmap pix, mask;
    int depth, argc;
    Position hotx, hoty;
    Arg args[8];
    XmDragOverBlend cblend = &DO_CursorBlend(w);

    if (op != NULL) {
	compute_pos(w, op, src, &sox, &soy);
	if (sox < x)
	    x = sox;
	if (soy < y)
	    y = soy;
    }

    if (state != NULL) {
	if (DI_Attachment(state) == XmATTACH_HOT && op != NULL) {
	    stx = sox + DI_HotX(op) - DI_HotX(state);
	    sty = soy + DI_HotY(op) - DI_HotY(state);
	}
	else
	    compute_pos(w, state, src, &stx, &sty);

	if (stx < x)
	    x = stx;
	if (sty < y)
	    y = sty;
    }

    cw = -x;
    ch = -y;
    lw = cw + DI_Width(src);
    lh = ch + DI_Height(src);

    if (op != NULL) {
	sox -= x;
	soy -= y;
	if (sox + DI_Width(op) > lw)
	    lw = sox + DI_Width(op);
	if (soy + DI_Height(op) > lh)
	    lh = soy + DI_Height(op);
	hotx = sox + DI_HotX(op);
	hoty = soy + DI_HotY(op);
    }
    else {
	hotx = cw + DI_HotX(src);
	hoty = cw + DI_HotY(src);
    }

    if (state != NULL) {
	stx -= x;
	sty -= y;
	if (stx + DI_Width(state) > lw)
	    lw = stx + DI_Width(state);
	if (sty + DI_Height(op) > lh)
	    lh = sty + DI_Height(state);
    }

    _XmGetMaxCursorSize(w, &sw, &sh);

    if (hotx > sw) {
	if (hotx > x - (sw >> 1))
	    hotx = sw >> 1;
	else
	    hotx = x - (sw >> 1);
    }
    else
	x = 0;

    if (hoty > sh) {
	if (hoty > y - (sh >> 1))
	    hoty = sh >> 1;
	else
	    hoty = y - (sh >> 1);
    }
    else
	y = 0;

    stx -= x;
    cw -= x;
    sox -= x;
    hotx -= x;
    sty -= y;
    ch -= y;
    if (ch > DI_HotY(src))
	ch = DI_HotY(src);
    soy -= y;
    hoty -= y;

    if (cblend == blend)
	depth = 1;
    else
	depth = CoreDepth(w);

#if 1
    lw = lh = 48;
#else
printf("DIM: %d %d %d %d %d %d %d %d %d %d\n", lw, lh, stx, sty, sox, soy, cw, ch, hotx, hoty);
#endif
    pix = _XmAllocScratchPixmap(s, depth, lw, lh);
#if 1
    mask = _XmAllocScratchPixmap(s, 1, lw, lh);
#else
    mask = XmUNSPECIFIED_PIXMAP;
#endif

    argc = 0;
    XtSetArg(args[argc], XmNpixmap, pix); argc++;
    XtSetArg(args[argc], XmNmask, mask); argc++;
    XtSetArg(args[argc], XmNdepth, depth); argc++;
    XtSetArg(args[argc], XmNwidth, lw); argc++;
    XtSetArg(args[argc], XmNheight, lh); argc++;
    XtSetArg(args[argc], XmNhotX, hotx); argc++;
    XtSetArg(args[argc], XmNhotY, hoty); argc++;

    blend->mixedIcon = (XmDragIconObject)XmCreateDragIcon(w, "mixedIcon",
							  args, argc);
    DI_XOffset(src) = cw;
    DI_YOffset(src) = ch;
    DI_XOffset(op) = sox;
    DI_YOffset(op) = soy;
    DI_XOffset(state) = stx;
    DI_YOffset(state) = sty;

    blend_icons(w, blend, src, op, state);
}

/*
 * create a pixmap cursor combining the icon objects
 */
static Cursor
get_cursor(Widget w,
	   XmDragIconObject src, XmDragIconObject op, XmDragIconObject state,
	   Boolean check_fit, Boolean dirty)
{
    Screen *scr = XtScreen(w);
    Display *disp = XtDisplay(w);
    Cursor ret = None;
    XColor backfore[2];
    XmDragCursorCache *cache = NULL, tmp = NULL;
    Pixmap pix, mask;
    Boolean in_cache = True;

    if (check_fit && !cursor_fits(w, src, op, state))
	return None;

    backfore[0].pixel = DO_CursorBackground(w);
    backfore[1].pixel = DO_CursorForeground(w);
    XQueryColors(disp, CoreColormap(w), backfore, 2);


    if (op == NULL || DI_Attachment(op) != XmATTACH_HOT) {

	cache = _XmGetDragCursorCachePtr((XmScreen)XmGetXmScreen(scr));
	if (*cache != NULL) {

	    tmp = *cache;

	    while (tmp) {

		if (tmp->sourceIcon == src && tmp->opIcon == op &&
		    tmp->stateIcon == state) {

		    if (dirty)
			break;

		    XRecolorCursor(disp, tmp->cursor,
				   &backfore[1], &backfore[0]);

		    return tmp->cursor;
		}

		tmp = tmp->next;
	    }
	}
    }
    else
	in_cache = False;


    make_soup(w, src, op, state, &DO_CursorBlend(w), check_fit);

    if (DI_Mask(DO_CursorBlend(w).mixedIcon) != XmUNSPECIFIED_PIXMAP)
	mask = DI_Mask(DO_CursorBlend(w).mixedIcon);
    else
	mask = None;

    pix = DI_Pixmap(DO_CursorBlend(w).mixedIcon);

    ret = XCreatePixmapCursor(disp, pix, mask, &backfore[1], &backfore[0],
			      DI_HotX(DO_CursorBlend(w).mixedIcon),
			      DI_HotY(DO_CursorBlend(w).mixedIcon));

    /* if it should be cached, update it there.  The caller will see that
     * it gets put into the DO_ActiveCursor field */
    if (in_cache) {

	if (tmp)
	    XFreeCursor(disp, tmp->cursor);
	else {
	    tmp = (XmDragCursorCache)XtMalloc(sizeof(XmDragCursorRec));
	    tmp->next = *cache;
	    tmp->cursor = ret;
	    tmp->stateIcon = state;
	    tmp->opIcon = op;
	    tmp->sourceIcon = src;
	    *cache = tmp;
	}

	tmp->cursor = ret;
    }
    /* otherwise, kill the current active cursor and replace it with this one */
    else {
	if (DO_ActiveCursor(w) != None)
	    XFreeCursor(disp, DO_ActiveCursor(w));

	DO_ActiveCursor(w) = ret;
    }

    return ret;
}

/*
 * (possibly) allocate colors for the cursor
 */
static Boolean
get_colors(Widget w)
{
    Widget par = XtParent(w); /* This had BETTER be a DragContext */
    Screen *scr = XtScreen(w);
    Display *disp = XtDisplay(w);
    XColor backfore[2];
    XGCValues values;
    Boolean new_gc = False;

    backfore[0].pixel = DC_CursorBackground(par);

    switch (DO_CursorState(w)) {
    case XmNO_DROP_SITE:
	backfore[1].pixel = DC_NoneCursorForeground(par);
	break;

    case XmINVALID_DROP_SITE:
	backfore[1].pixel = DC_InvalidCursorForeground(par);
	break;

    case XmVALID_DROP_SITE:
	backfore[1].pixel = DC_ValidCursorForeground(par);
	break;

    default:
	_XmWarning(w, "DragCursor in invalid state.");
	backfore[1].pixel = DC_NoneCursorForeground(par);
	break;
    }

    XQueryColors(disp, CoreColormap(w), backfore, 2);

    if (!XAllocColor(disp, DefaultColormapOfScreen(scr), &backfore[0]) ||
	!XAllocColor(disp, DefaultColormapOfScreen(scr), &backfore[1]) ||
	backfore[0].pixel == backfore[1].pixel) {
	backfore[0].pixel = BlackPixelOfScreen(scr);
	backfore[1].pixel = WhitePixelOfScreen(scr);
    }

    /* root GC not created */
    if (DO_RootBlend(w).gc == NULL) {

	DO_CursorBackground(w) = backfore[0].pixel;
	DO_CursorForeground(w) = backfore[1].pixel;

	values.foreground = backfore[1].pixel;
	values.background = backfore[0].pixel;
	/* just like PanedWindow for line effects. */
	values.subwindow_mode = IncludeInferiors;
	values.graphics_exposures = False;

	DO_RootBlend(w).gc = XCreateGC(disp, RootWindowOfScreen(scr),
				       GCForeground|GCBackground|
					GCSubwindowMode|GCGraphicsExposures,
				       &values);
	new_gc = True;
    }
    /* root GC already created?  Update fg and bg */
    else {

	DO_CursorBackground(w) = backfore[0].pixel;
	DO_CursorForeground(w) = backfore[1].pixel;

	values.foreground = backfore[1].pixel;
	values.background = backfore[0].pixel;

	XChangeGC(disp, DO_RootBlend(w).gc, GCForeground|GCBackground,
		  &values);
    }

    return new_gc;
}

/*
 * destroy the blended cursor
 */
static void
destroy_mixed(Widget w, XmDragIconObject icon)
{
    XmScreen scr = (XmScreen)XmGetXmScreen(XtScreen(w));

    if (DI_Pixmap(icon)) {
	_XmFreeScratchPixmap(scr, DI_Pixmap(icon));
	DI_Pixmap(icon) = XmUNSPECIFIED_PIXMAP;
    }

    if (DI_Mask(icon)) {
	_XmFreeScratchPixmap(scr, DI_Mask(icon));
	DI_Pixmap(icon) = XmUNSPECIFIED_PIXMAP;
    }

    XtDestroyWidget((Widget)icon);
}

static void
change_active_mode(Widget w)
{
}

static void
change_window(Widget w)
{
}

/*
 * melt the icon
 */
static void
melt_cursor(Widget w)
{
}

/*
 * return the icon back to the drag site
 */
static void
zap_cursor(Widget w)
{
}

void
_XmDragOverHide(Widget w, Position clipOriginX, Position clipOriginY,
		XmRegion clipRegion)
{
    XtPopdown(w);
}

void
_XmDragOverShow(Widget w, Position clipOriginX, Position clipOriginY,
		XmRegion clipRegion)
{
    XtPopup(w, XtGrabNonexclusive);
}

void
_XmDragOverMove(Widget w, Position x, Position y)
{
}

void
_XmDragOverChange(Widget w, unsigned char dropSiteStatus)
{
    Widget par = XtParent(w); /* This had BETTER be a DragContext */
    XmDragIconObject src_icon = NULL, op_icon = NULL, state_icon = NULL;
    Boolean is_root = True, dirty, change;

    DO_CursorState(w) = dropSiteStatus;
    if (DO_Mode(w) == XmWINDOW)
	return;

    if (DC_BlendModel(par) == XmBLEND_NONE)
	return;

    if (DO_Mode(w) == XmPIXMAP)
	src_icon = DC_SourcePixmapIcon(par);

    if (src_icon == NULL || XtScreen(w) != XtScreen(src_icon) ||
	(DI_Depth(src_icon) != CoreDepth(w) && DI_Depth(src_icon) != 1)) {

	is_root = False;

	src_icon = DC_SourceCursorIcon(par);

	if (src_icon == NULL || XtScreen(w) != XtScreen(src_icon) ||
	    (DI_Depth(src_icon) != CoreDepth(w) && DI_Depth(src_icon) != 1))
	    src_icon = _XmScreenGetSourceIcon(w);
    }

    /*
     * gather up the cursors according to blendModel
     */
    switch (DC_BlendModel(par)) {
    default:
	_XmWarning(w, "Invalid blend model.");
	/* FALL THROUGH (i.e., if they screw up, give them everything */

    case XmBLEND_ALL:

	op_icon = DC_OperationCursorIcon(par);

	if (op_icon == NULL || XtScreen(w) != XtScreen(op_icon) ||
	    DI_Depth(op_icon) != 1) {

	    op_icon = _XmScreenGetOperationIcon(w, DC_Operation(par));

	    if (op_icon && DI_Depth(op_icon) != 1)
		op_icon = NULL;
	}
	/* FALL THROUGH */

    case XmBLEND_STATE_SOURCE:

	state_icon = DC_StateCursorIcon(par);

	if (state_icon == NULL || XtScreen(w) != XtScreen(state_icon) ||
	    DI_Depth(state_icon) != 1) {

	    state_icon = _XmScreenGetStateIcon(w, dropSiteStatus);

	    if (state_icon && DI_Depth(state_icon) != 1)
		state_icon = NULL;
	}
	/* FALL THROUGH */

    case XmBLEND_JUST_SOURCE:

	if (!_XmDragIconIsDirty(src_icon) &&
	    (op_icon == NULL || !_XmDragIconIsDirty(op_icon)) &&
	    (state_icon == NULL || !_XmDragIconIsDirty(state_icon)))
	    dirty = False;
	else
	    dirty = True;
    }

    if (!get_colors(w) &&
	DO_StateIcon(w) == state_icon &&
	DO_OpIcon(w) == op_icon &&
	DO_RootBlend(w).sourceIcon == src_icon &&
	!dirty)
	change = False;
    else
	change = True;

    DO_StateIcon(w) = state_icon;
    DO_OpIcon(w) = op_icon;
    DO_CursorBlend(w).sourceIcon = src_icon;

    if (!is_root &&
	(DO_ActiveCursor(w) = get_cursor(w, src_icon, op_icon, state_icon,
					 False, dirty)) != None) {

	_XmDragIconClean(src_icon, state_icon, op_icon);

	if (DO_ActiveMode(w) != XmCURSOR) {
	    _XmDragOverHide(w, 0, 0, NULL);
	    DO_ActiveMode(w) = XmCURSOR;
	}

	XChangeActivePointerGrab(XtDisplay(w),
				 ButtonPressMask|ButtonReleaseMask|
				 ButtonMotionMask|
				 (!DC_TrackingMode(par)
					? (EnterWindowMask|LeaveWindowMask)
					: 0),
				 DO_ActiveCursor(w),
				 DC_LastChangeTime(par));

	DO_RootBlend(w).sourceIcon = NULL;
	if (DO_RootBlend(w).mixedIcon)
	    destroy_mixed(w, DO_RootBlend(w).mixedIcon);
    }
    else if (DO_Mode(w) == XmCURSOR) {
	DO_RootBlend(w).sourceIcon = src_icon;

	DO_ActiveCursor(w) = get_cursor(w, src_icon, op_icon, state_icon,
					False, dirty);

	_XmDragIconClean(src_icon, state_icon, op_icon);

	if (DO_ActiveMode(w) != XmCURSOR) {
	    _XmDragOverHide(w, 0, 0, NULL);
	    DO_ActiveMode(w) = XmCURSOR;
	}

	XChangeActivePointerGrab(XtDisplay(w),
				 ButtonPressMask|ButtonReleaseMask|
				 ButtonMotionMask|
				 (!DC_TrackingMode(par)
					? (EnterWindowMask|LeaveWindowMask)
					: 0),
				 DO_ActiveCursor(w),
				 DC_LastChangeTime(par));
    }
    else {

	DO_RootBlend(w).sourceIcon = src_icon;

	if (change || DO_ActiveMode(w) != XmPIXMAP) {
	    _XmDragIconClean(src_icon, state_icon, op_icon);

	    change_active_mode(w);
	}
    }
}

/*
 * called on BRelease
 */
void
_XmDragOverFinish(Widget w, unsigned char completionStatus)
{
    Widget dc = XtParent(w); /* this had BETTER be a drag context */
    Display *disp = XtDisplay(w);

    if (DC_BlendModel(dc) == XmBLEND_NONE)
	return;

    XGrabServer(disp);

    change_window(w);

    if (completionStatus != XmDROP)
	melt_cursor(w);
    else
	zap_cursor(w);

    XtPopdown(w);

    DO_IsVisible(w) = False;

    XUngrabServer(disp);
}

Cursor
_XmDragOverGetActiveCursor(Widget w)
{
    return DO_ActiveCursor(w);
}

void
_XmDragOverSetInitialPosition(Widget w, Position initialX, Position initialY)
{
    DO_InitialX(w) = initialX;
    DO_InitialY(w) = initialY;
}

