/**
 *
 * $Id: DropSMgr.c,v 1.8 1997/02/09 04:40:34 miers Exp $
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

static char rcsid[] = "$Id: DropSMgr.c,v 1.8 1997/02/09 04:40:34 miers Exp $";

#include <LTconfig.h>
#include <XmI/XmI.h>

#include <Xm/DropSMgrP.h>
#include <Xm/DisplayP.h>
#include <XmI/DragDropI.h>
#include <Xm/DragCP.h>
#include <Xm/XmP.h>

#include <XmI/DebugUtil.h>

#ifndef XmUNSPECIFIED
#define XmUNSPECIFIED (~0)
#endif

/* Forward Declarations */

static void __XmDSMCreateInfo(XmDropSiteManagerObject dsm,
			      Widget widget,
			      ArgList args,
			      Cardinal num_args);
static void __XmDSMDestroyInfo(XmDropSiteManagerObject dsm,
			       Widget widget);
static void __XmDSMStartUpdate(XmDropSiteManagerObject dsm,
			       Widget widget);
static void __XmDSMRetrieveInfo(XmDropSiteManagerObject dsm,
				Widget widget,
				ArgList args,
				Cardinal num_args);
static void __XmDSMUpdateInfo(XmDropSiteManagerObject dsm,
			      Widget widget,
			      ArgList args,
			      Cardinal num_args);
static void __XmDSMEndUpdate(XmDropSiteManagerObject dsm,
			     Widget widget);
static void __XmDSMUpdateDSM(XmDropSiteManagerObject dsm,
			     XtPointer clientData,
			     XtPointer callData);
static void __XmDSMProcessMotion(XmDropSiteManagerObject dsm,
				 XtPointer clientData,
				 XtPointer callData);
static void __XmDSMProcessDrop(XmDropSiteManagerObject dsm,
			       XtPointer clientData,
			       XtPointer callData);
static void __XmDSMOperationChanged(XmDropSiteManagerObject dsm,
				    XtPointer clientData,
				    XtPointer callData);
static void __XmDSMChangeRoot(XmDropSiteManagerObject dsm,
			      XtPointer clientData,
			      XtPointer callData);
static void __XmDSMInsertInfo(XmDropSiteManagerObject dsm,
			      XtPointer clientData,
			      XtPointer callData);
static void __XmDSMRemoveInfo(XmDropSiteManagerObject dsm,
			      XtPointer info);
static void __XmDSMSyncTree(XmDropSiteManagerObject dsm,
			    Widget shell);
static int __XmDSMGetTree(XmDropSiteManagerObject dsm,
			  Widget shell,
			  XtPointer dataPtr);
static void __XmDSMCreateDSInfoTable(XmDropSiteManagerObject dsm);
static void __XmDSMDestroyDSInfoTable(XmDropSiteManagerObject dsm);
static void __XmDSMRegisterInfo(XmDropSiteManagerObject dsm,
				Widget widget,
				XtPointer info);
static XtPointer __XmDSMWidgetToInfo(XmDropSiteManagerObject dsm,
				     Widget widget);
static void __XmDSMUnregisterInfo(XmDropSiteManagerObject dsm,
				  XtPointer info);

static void class_initialize(void);
static void class_part_initialize(WidgetClass widget_class);
static void initialize(Widget request, Widget new_w, ArgList args, Cardinal *num_args);
static void destroy(Widget w);
static Boolean set_values(Widget current, Widget request, Widget new_w, ArgList args, Cardinal *num_args);

#define Offset(f)	XtOffsetOf(XmDropSiteManagerRec, dropManager.f)
static XtResource resources[] = {
    {
	XmNnotifyProc, XmCNotifyProc, XmRCallbackProc,
	sizeof(XtCallbackList), Offset(notifyProc),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNtreeUpdateProc, XmCTreeUpdateProc, XmRCallbackProc,
	sizeof(XtCallbackList), Offset(treeUpdateProc),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNclientData, XmCClientData, XmRPointer,
	sizeof(XtPointer), Offset(client_data),
	XmRImmediate, (XtPointer)NULL
    }
};
#undef Offset

/*
 * these are documented as belonging here.  Where do they go?
 */
typedef struct {
    unsigned char	dropSiteOperations;
    XtCallbackProc	dragProc;
    XtCallbackProc	dropProc;
    unsigned char	dropSiteType;
    unsigned char	animationStyle;
    unsigned char	dropSiteActivity;
    Atom		*importTargets;
    Cardinal		numImportTargets;
    XRectangle		dropRectangles;
    Cardinal		numDropRectangles;
    Pixmap		animationPixmap;
    Cardinal		animationPixmapDepth;
    Pixmap		animationMask;
} XmDropSiteHiddenResources;

#define Offset(f)	XtOffsetOf(XmDropSiteHiddenResources, f)
static XtResource hidden_resources[] = {
    {
	XmNanimationMask, XmCAnimationMask, XmRAnimationMask,
	sizeof(Pixmap), Offset(animationMask),
	XmRImmediate, (XtPointer)XmUNSPECIFIED_PIXMAP
    },
    {
	XmNanimationPixmap, XmCAnimationPixmap, XmRAnimationPixmap,
	sizeof(Pixmap), Offset(animationPixmap),
	XmRImmediate, (XtPointer)XmUNSPECIFIED_PIXMAP
    },
    {
	XmNanimationPixmapDepth, XmCAnimationPixmapDepth, XmRCardinal,
	sizeof(Cardinal), Offset(animationPixmapDepth),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNanimationStyle, XmCAnimationStyle, XmRAnimationStyle,
	sizeof(unsigned char), Offset(animationStyle),
	XmRImmediate, (XtPointer)XmDRAG_UNDER_HIGHLIGHT
    },
    {
	XmNdragProc, XmCDragProc, XmRProc,
	sizeof(XtCallbackProc), Offset(dragProc),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNdropProc, XmCDropProc, XmRProc,
	sizeof(XtCallbackProc), Offset(dropProc),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNdropRectangles, XmCDropRectangles, XmRRectangleList,
	sizeof(XRectangle *), Offset(dropRectangles),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNdropSiteActivity, XmCDropSiteActivity, XmRDropSiteActivity,
	sizeof(unsigned char), Offset(dropSiteActivity),
	XmRImmediate, (XtPointer)XmDROP_SITE_ACTIVE
    },
    {
	XmNdropSiteType, XmCDropSiteType, XmRDropSiteType,
	sizeof(unsigned char), Offset(dropSiteType),
	XmRImmediate, (XtPointer)XmDROP_SITE_SIMPLE
    },
    {
	XmNimportTargets, XmCImportTargets, XmRAtomList,
	sizeof(Atom *), Offset(importTargets),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNnumImportTargets, XmCNumImportTargets, XmRCardinal,
	sizeof(Cardinal), Offset(numImportTargets),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNnumDropRectangles, XmCNumDropRectangles, XmRCardinal,
	sizeof(Cardinal), Offset(numDropRectangles),
	XmRImmediate, (XtPointer)1
    },
    {
	XmNdropSiteOperations, XmCDropSiteOperations, XmRDropSiteOperations,
	sizeof(unsigned char), Offset(dropSiteOperations),
	XmRImmediate, (XtPointer)(XmDROP_MOVE|XmDROP_COPY)
    }
};
#undef Offset

static XmBaseClassExtRec _XmDropSMgrObjectClassExtRec = {
    /* next_extension            */ NULL,
    /* record_type               */ NULLQUARK,                             
    /* version                   */ XmBaseClassExtVersion,
    /* size                      */ sizeof(XmBaseClassExtRec),
    /* initialize_prehook        */ NULL, /* FIXME */
    /* set_values_prehook        */ NULL, /* FIXME */
    /* initialize_posthook       */ NULL, /* FIXME */
    /* set_values_posthook       */ NULL, /* FIXME */
    /* secondary_object_class    */ NULL, /* FIXME */
    /* secondary_object_create   */ NULL, /* FIXME */
    /* get_secondary_resources   */ NULL, /* FIXME */
    /* fast_subclass             */ { 0 }, /* FIXME */
    /* get_values_prehook        */ NULL, /* FIXME */
    /* get_values_posthook       */ NULL, /* FIXME */
    /* class_part_init_prehook   */ NULL, /* FIXME */
    /* class_part_init_posthook  */ NULL, /* FIXME */
    /* ext_resources             */ NULL, /* FIXME */
    /* compiled_ext_resources    */ NULL, /* FIXME */
    /* num_ext_resources         */ 0, /* FIXME */
    /* use_sub_resources         */ False, /* FIXME */
    /* widget_navigable          */ NULL, /* FIXME */
    /* focus_change              */ NULL, /* FIXME */
    /* wrapper_data              */ NULL
};

XmDropSiteManagerClassRec xmDropSiteManagerClassRec = {
    /* Object class part */
    {
	/* superclass            */ (WidgetClass) &objectClassRec,
        /* class_name            */ "XmDropSiteManager",
	/* widget_size           */ sizeof(XmDropSiteManagerRec),
	/* class_initialize      */ class_initialize,
	/* class_part_initialize */ class_part_initialize,
	/* class_inited          */ False,
	/* initialize            */ initialize,
	/* initialize_hook       */ NULL,
	/* obj1                  */ NULL,
	/* obj2                  */ NULL,
	/* obj3                  */ 0,
	/* resources             */ resources,
	/* num_resources         */ XtNumber(resources),
	/* xrm_class             */ NULLQUARK,
	/* obj4                  */ 0,
	/* obj5                  */ 0,
	/* obj6                  */ 0,
	/* obj7                  */ 0,
	/* destroy               */ destroy,
	/* obj8                  */ NULL,
	/* obj9                  */ NULL,
	/* set_values            */ set_values,
	/* set_values_hook       */ NULL,
	/* obj10                 */ NULL,
	/* get_values_hook       */ NULL,
	/* obj11                 */ NULL,
	/* version               */ XtVersion,
	/* callback offsets      */ NULL,
	/* obj12                 */ NULL,
	/* obj13                 */ NULL,
	/* obj14                 */ NULL,
	/* extension             */ (XtPointer)&_XmDropSMgrObjectClassExtRec
    },
    /* Drop site manager class part */
    {
	/* createInfo       */  __XmDSMCreateInfo,
	/* destroyInfo      */  __XmDSMDestroyInfo,
	/* startUpdate      */  __XmDSMStartUpdate,
	/* retrieveInfo     */  __XmDSMRetrieveInfo,
	/* updateInfo       */  __XmDSMUpdateInfo,
	/* endUpdate        */  __XmDSMEndUpdate,
	/* updateDSM        */  __XmDSMUpdateDSM,
	/* processMotion    */  __XmDSMProcessMotion,
	/* processDrop      */  __XmDSMProcessDrop,
	/* operationChanged */  __XmDSMOperationChanged,
	/* changeRoot       */  __XmDSMChangeRoot,
	/* insertInfo       */  __XmDSMInsertInfo,
	/* removeInfo       */  __XmDSMRemoveInfo,
	/* syncTree         */  __XmDSMSyncTree,
	/* getTreeFromDSM   */  __XmDSMGetTree,
	/* createTable      */  __XmDSMCreateDSInfoTable,
	/* destroyTable     */  __XmDSMDestroyDSInfoTable,
	/* registerInfo     */  __XmDSMRegisterInfo,
	/* widgetToInfo     */  __XmDSMWidgetToInfo,
	/* unregisterInfo   */  __XmDSMUnregisterInfo,
	/* extension        */  NULL
    }
};

WidgetClass xmDropSiteManagerObjectClass = (WidgetClass)(&xmDropSiteManagerClassRec);

/*
 * Creates a drop site for the associated widget (perhaps)
 */

static void
class_initialize(void)
{
    _XmDropSMgrObjectClassExtRec.record_type = XmQmotif;
}

static void
class_part_initialize(WidgetClass widget_class)
{
    WidgetClass sc;

    _XmFastSubclassInit(widget_class, XmDROP_SITE_MANAGER_BIT);

    sc = widget_class->core_class.superclass;

    if (DSC_CreateInfoProc(widget_class) == XmInheritCreateInfoProc)
	DSC_CreateInfoProc(widget_class) = DSC_CreateInfoProc(sc);

    if (DSC_DestroyInfoProc(widget_class) == XmInheritDestroyInfoProc)
	DSC_DestroyInfoProc(widget_class) = DSC_DestroyInfoProc(sc);

    if (DSC_StartUpdateProc(widget_class) == XmInheritStartUpdateProc)
	DSC_StartUpdateProc(widget_class) = DSC_StartUpdateProc(sc);

    if (DSC_RetrieveInfoProc(widget_class) == XmInheritRetrieveInfoProc)
	DSC_RetrieveInfoProc(widget_class) = DSC_RetrieveInfoProc(sc);

    if (DSC_UpdateInfoProc(widget_class) == XmInheritUpdateInfoProc)
	DSC_UpdateInfoProc(widget_class) = DSC_UpdateInfoProc(sc);

    if (DSC_EndUpdateProc(widget_class) == XmInheritEndUpdateProc)
	DSC_EndUpdateProc(widget_class) = DSC_EndUpdateProc(sc);

    if (DSC_UpdateProc(widget_class) == XmInheritUpdateProc)
	DSC_UpdateProc(widget_class) = DSC_UpdateProc(sc);

    if (DSC_ProcessMotionProc(widget_class) == XmInheritProcessMotionProc)
	DSC_ProcessMotionProc(widget_class) = DSC_ProcessMotionProc(sc);

    if (DSC_ProcessDropProc(widget_class) == XmInheritProcessDropProc)
	DSC_ProcessDropProc(widget_class) = DSC_ProcessDropProc(sc);

    if (DSC_OperationChangedProc(widget_class) == XmInheritOperationChangedProc)
	DSC_OperationChangedProc(widget_class) = DSC_OperationChangedProc(sc);

    if (DSC_ChangeRootProc(widget_class) == XmInheritChangeRootProc)
	DSC_ChangeRootProc(widget_class) = DSC_ChangeRootProc(sc);

    if (DSC_InsertInfoProc(widget_class) == XmInheritInsertInfoProc)
	DSC_InsertInfoProc(widget_class) = DSC_InsertInfoProc(sc);

    if (DSC_RemoveInfoProc(widget_class) == XmInheritRemoveInfoProc)
	DSC_RemoveInfoProc(widget_class) = DSC_RemoveInfoProc(sc);

    if (DSC_SyncTreeProc(widget_class) == XmInheritSyncTreeProc)
	DSC_SyncTreeProc(widget_class) = DSC_SyncTreeProc(sc);

    if (DSC_GetTreeFromDSMProc(widget_class) == XmInheritGetTreeFromDSMProc)
	DSC_GetTreeFromDSMProc(widget_class) = DSC_GetTreeFromDSMProc(sc);

    if (DSC_CreateDSInfoTable(widget_class) == XmInheritCreateDSInfoTable)
	DSC_CreateDSInfoTable(widget_class) = DSC_CreateDSInfoTable(sc);

    if (DSC_DestroyDSInfoTable(widget_class) == XmInheritDestroyDSInfoTable)
	DSC_DestroyDSInfoTable(widget_class) = DSC_DestroyDSInfoTable(sc);

    if (DSC_RegisterInfoProc(widget_class) == XmInheritRegisterInfoProc)
	DSC_RegisterInfoProc(widget_class) = DSC_RegisterInfoProc(sc);

    if (DSC_WidgetToInfoProc(widget_class) == XmInheritWidgetToInfoProc)
	DSC_WidgetToInfoProc(widget_class) = DSC_WidgetToInfoProc(sc);

    if (DSC_UnregisterInfoProc(widget_class) == XmInheritUnregisterInfoProc)
	DSC_UnregisterInfoProc(widget_class) = DSC_UnregisterInfoProc(sc);
}

static void
initialize(Widget request, Widget new_w, ArgList args, Cardinal *num_args)
{
    DS_CurAnimate(new_w) = True;

    DS_CurDropSiteStatus(new_w) = 2;
    DS_CurOperations(new_w) = 0;
    DS_CurOperation(new_w) = 0;

    DS_CurX(new_w) = 0;
    DS_CurY(new_w) = 0;
    DS_OldX(new_w) = 0;
    DS_OldY(new_w) = 0;

    DS_CurDragContext(new_w) = NULL;
    DS_CurTime(new_w) = CurrentTime;
    DS_CurInfo(new_w) = NULL;
    DS_DragUnderData(new_w) = NULL;

    DS_CurAncestorClipRegion(new_w) = _XmRegionCreate();
    DS_NewAncestorClipRegion(new_w) = _XmRegionCreate();

    DSMCreateTable((XmDropSiteManagerObject)new_w);

    DS_RootX(new_w) = 0;
    DS_RootY(new_w) = 0;

    DS_RootWidth(new_w) = XmUNSPECIFIED;
    DS_RootHeight(new_w) = XmUNSPECIFIED;

    DS_DSRoot(new_w) = NULL;
    DS_ClipperList(new_w) = NULL;
    DS_UpdateInfo(new_w) = NULL;
}

static void
destroy(Widget w)
{
    DSMDestroyTable((XmDropSiteManagerObject)w);

    _XmRegionDestroy(DS_CurAncestorClipRegion(w));
    _XmRegionDestroy(DS_NewAncestorClipRegion(w));
}

static Boolean
set_values(Widget current, Widget request, Widget new_w,
	   ArgList args, Cardinal *num_args)
{
    return True;
}

/*
 * looks like it calls:
 * _XmInitTargetsTable
 * adds _XmIEndUpdate as an AppTimeOut
 */
static void 
__XmDSMCreateInfo(XmDropSiteManagerObject object,
		  Widget widget,
		  ArgList args,
		  Cardinal num_args)
{
    /* Fix Me */
}

static void 
__XmDSMDestroyInfo(XmDropSiteManagerObject dsm,
		   Widget widget)
{
}

static void 
__XmDSMStartUpdate(XmDropSiteManagerObject dsm,
		   Widget widget)
{
}

static void 
__XmDSMRetrieveInfo(XmDropSiteManagerObject dsm,
		    Widget widget,
		    ArgList args,
		    Cardinal num_args)
{
}

static void 
__XmDSMUpdateInfo(XmDropSiteManagerObject dsm,
		  Widget widget,
		  ArgList args,
		  Cardinal num_args)
{
}

static void 
__XmDSMEndUpdate(XmDropSiteManagerObject dsm,
		 Widget widget)
{
}

static void 
__XmDSMUpdateDSM(XmDropSiteManagerObject dsm,
		 XtPointer clientData,
		 XtPointer callData)
{
}

/*
 * When a drag motion event is sent to this client, we handle 
 * the notification of the drop site (if it has drag under effects)
 */
static void 
__XmDSMProcessMotion(XmDropSiteManagerObject dsm,
		     XtPointer clientData,
		     XtPointer callData)
{
}

/*
 * When the drop is started (BDrag is released), we let the drop site know that
 * it's being dropped in
 */
static void 
__XmDSMProcessDrop(XmDropSiteManagerObject dsm,
		   XtPointer clientData,
		   XtPointer callData)
{
}

/*
 * We notify the current drop site (if any) that the operation of the drag
 * has changed
 */
static void 
__XmDSMOperationChanged(XmDropSiteManagerObject dsm,
			XtPointer clientData,
			XtPointer callData)
{
}

static void 
__XmDSMChangeRoot(XmDropSiteManagerObject dsm,
		  XtPointer clientData,
		  XtPointer callData)
{
}

static void 
__XmDSMInsertInfo(XmDropSiteManagerObject dsm,
		  XtPointer clientData,
		  XtPointer callData)
{
}

static void 
__XmDSMRemoveInfo(XmDropSiteManagerObject dsm,
		  XtPointer info)
{
}

static void 
__XmDSMSyncTree(XmDropSiteManagerObject dsm,
		Widget shell)
{
}

/*
 * This function creates (or replaces) the property on the shell widget
 * to contain the entire drop site database.
 */
static int 
__XmDSMGetTree(XmDropSiteManagerObject dsm,
	       Widget shell,
	       XtPointer dataPtr)
{
    return 0;
}

/*
 * Creates a hash table that goes from widgets to drop site information
 */
static void 
__XmDSMCreateDSInfoTable(XmDropSiteManagerObject dsm)
{
}

/*
 * destroys the hash table
 */
static void 
__XmDSMDestroyDSInfoTable(XmDropSiteManagerObject dsm)
{
}

/*
 * adds a widget and its associated info to the hash table
 */
static void 
__XmDSMRegisterInfo(XmDropSiteManagerObject dsm,
		    Widget widget,
		    XtPointer info)
{
}

/*
 * retrieves the info for a given widget from the hash table
 */
static XtPointer 
__XmDSMWidgetToInfo(XmDropSiteManagerObject dsm,
		    Widget widget)
{
    return NULL;
}

/*
 * removes the info from the hash table, along with the widget that
 * hashes to it.
 */
static void 
__XmDSMUnregisterInfo(XmDropSiteManagerObject dsm,
		      XtPointer info)
{
}

void
_XmDSMUpdate(XmDropSiteManagerObject dsm,
	     XtPointer clientData,
	     XtPointer callData)
{
    DSMUpdate(dsm, clientData, callData);
}

int
_XmDSMGetTreeFromDSM(XmDropSiteManagerObject dsm,
		     Widget shell,
		     XtPointer dataPtr)
{
    return DSMGetTreeFromDSM(dsm, shell, dataPtr);
}

Boolean
_XmDropSiteShell(Widget widget)
{
    Widget disp = XmGetXmDisplay(XtDisplay(widget));
    XmDropSiteManagerObject dsm;

    dsm = _XmGetDropSiteManagerObject((XmDisplay)disp);

    if (!XtIsShell(widget))
	return False;

    if (DSMWidgetToInfo(dsm, widget))
	return True;

    return False;
}

static Boolean
has_ds_offspring(XmDropSiteManagerObject dsm, Widget w)
{
    int i;

    if (!XtIsComposite(w))
	return False;

    for (i = 0; i < MGR_NumChildren(w); i++) {

	if (DSMWidgetToInfo(dsm, w))
	    return True;

	if (has_ds_offspring(dsm, MGR_Children(w)[i]))
	    return True;
    }

    return False;
}

Boolean
_XmDropSiteWrapperCandidate(Widget widget)
{
    Widget disp = XmGetXmDisplay(XtDisplay(widget)), par;
    XmDropSiteManagerObject dsm;

    dsm = _XmGetDropSiteManagerObject((XmDisplay)disp);

    if (!dsm)
	return False;

    if (DSMWidgetToInfo(dsm, widget))
	return True;

    if (!XtIsComposite(widget))
	return False;

    par = widget;
    while (!XtIsShell(par))
	par = XtParent(par);

    if (!_XmDropSiteShell(par))
	return False;

    return has_ds_offspring(dsm, widget);
}

Widget
_XmGetActiveDropSite(Widget widget)
{
    Widget disp = XmGetXmDisplay(XtDisplay(widget));
    XmDropSiteManagerObject dsm;

    dsm = _XmGetDropSiteManagerObject((XmDisplay)disp);

    /* FIXME */
    return NULL;
}

void
_XmSyncDropSiteTree(Widget shell)
{
    Widget disp = XmGetXmDisplay(XtDisplay(shell));
    XmDropSiteManagerObject dsm;

    dsm = _XmGetDropSiteManagerObject((XmDisplay)disp);

    DSMSyncTree(dsm, shell);
}

/*
 * calls display's _XmGetProtocolStyle()
 */
void
_XmIEndUpdate(XtPointer client_data, XtIntervalId *interval_id)
{
    /* FIXME */
}

